package se.umu.jobsubmission.impl.datamanager; 

import java.io.File;
import java.io.FileNotFoundException; 
import java.io.RandomAccessFile;
import java.net.URI;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._06.jsdl.DataStaging_Type;
import org.ggf.schemas.jsdl._2005._06.jsdl.JobDefinition_Type;
import org.globus.ftp.DataChannelAuthentication;
import org.globus.ftp.FTPClient;
import org.globus.ftp.GridFTPClient;
import org.globus.ftp.GridFTPSession;
import org.globus.ftp.Marker;
import org.globus.ftp.MarkerListener;
import org.globus.ftp.extended.GridFTPControlChannel;
import org.globus.ftp.vanilla.Command;
import org.globus.ftp.vanilla.Reply;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.ietf.jgss.GSSCredential;

import se.umu.wsag.util.JobSubmissionUtils;

/**
 * 
 * @author tordsson
 *
 * Used for various data management tasks related to job submission. 
 */

public class DataHelper {
	
	private class DummyMarkerListener implements MarkerListener {

	    public DummyMarkerListener() {
	    }

	    public void markerArrived(Marker m) {
	    }

	}

    private static Logger logger = Logger.getLogger(DataHelper.class); 

    /**
     * Returns size of local file. 
     * @param fileName
     * @return
     * @throws Exception
     */
    private long getLocalSize(URI uri) throws Exception {
        String fileName = uri.getSchemeSpecificPart(); 
        logger.debug("looking up local size of: " + fileName); 
        File f = new File(fileName);
        return f.length(); 
    }

    /**
     * Returns size of file stored at GridFTP server. 
     * @param url
     * @return
     * @throws Exception
     */
    private long getGSIFTPSize(URI uri) throws Exception {

        String host = uri.getHost();
        int port = uri.getPort(); 	
        String fileName = uri.getPath(); 

        logger.debug("Getting size of [" + fileName + "]" 
                + " at [" + host + "] : [" + port + "]"); 

        GridFTPClient client = null; 
        GSSCredential cred = null; 

        // Get delegated credential. 
        GlobusCredential gCred = JobSubmissionUtils.getDelegatedCredential(); 
        cred = new GlobusGSSCredentialImpl(gCred, 
                GSSCredential.DEFAULT_LIFETIME); 

        // Open client connection, get size and close connection. 
        client = new GridFTPClient(host, port); 
        client.authenticate(cred); 
        long size = client.getSize(fileName); 
        client.close();

        return size; 
    }

    /**
     * Same as getGSIFTPSize, but connects to to an ftp server 
     * as user "anonymous". 
     * @param url
     * @return
     * @throws Exception
     */
    public long getFTPSize(URI uri) throws Exception {

        String host = uri.getHost(); 
        int port = uri.getPort(); 
        String fileName = uri.getSchemeSpecificPart(); 

        logger.debug("Getting size of [" + fileName + "]" 
                + " at [" + host + "] : [" + port + "]"); 

        FTPClient client = null; 
        client = new FTPClient(host, port); 
        client.authorize("anonymous", ""); // This is the preferred way?
        long size = client.size(fileName); 
        client.close(); 

        return size;
    }


    /**
     * Extracts list of job input files from job description. 
     * @param jobDescription
     * @return
     * @throws Exception
     */
    public List getInputFiles(JobDefinition_Type jobDescription) 
    throws Exception {

        List result = new LinkedList(); 
     
        // TODO: DataStaging can include both files to stage in and 
        // files to stage out...
        DataStaging_Type staging[] =  jobDescription.getJobDescription().
				getDataStaging(); 
        
        // TODO: Is this ok? 
        if(staging.length == 0)
        	return null; 
        
        for(int i = 0; i < staging.length; i++) {
        	logger.debug("Found file: " + staging[i].getSource().getURI().toString()); 
        	result.add(staging[i].getSource().getURI().toString()); 
        }

        logger.debug("The following are input files: " + result); 
        return result;     	
    }

    /**
     * Determines sizes of ftp (or gsiftp) files or local files. 
     * @param fileURI
     * @return
     * @throws Exception
     */
    public long getFileSize(URI fileURI) throws Exception {

        String protocol = fileURI.getScheme(); 

        logger.debug("Protocol is: " + protocol); 

        if("file".equalsIgnoreCase(protocol)) {
            return getLocalSize(fileURI); 
        }
        if("gsiftp".equalsIgnoreCase(protocol)) {
            return getGSIFTPSize(fileURI); 
        }
        if("ftp".equalsIgnoreCase(protocol)) {
            return getFTPSize(fileURI); 
        }

        logger.debug("Protocol " + protocol + " is not supported"); 				
        throw new Exception("Protocol " + protocol + " is not supported"); 						

    }

    /**
     * Determines the sizes in a list of inputfiles. 
     * Files can be stored locally (file://), or at a
     * FTP server (gsiftp:// or ftp://). 
     * A Map containing the (file, filesize) tuples is returned. 
     * @param fileNames
     * @return
     */
    public Map getFileSizes(List fileNames) {

        Map files = new HashMap(); 

        String fileName = null; 
        Long fileSize;
        long size; 
        URI uri; 
        for(int i = 0; i < fileNames.size(); i++) {
            try {
                fileName = (String) fileNames.get(i); 
                uri = new URI(fileName); 
                size = getFileSize(uri); 
                fileSize = new Long(size); 
                files.put(fileName, fileSize); 
            } 
            catch (Exception e) {
                logger.debug("Can not determine size of file " + fileName); 
                logger.debug("Ignoring it..."); 
            }						
        }

        return files; 

    }
    
    public void makeFileExecutable(String host, int port, String filePath, 
    		GSSCredential cred) throws Exception {
    	
    	GridFTPControlChannel ctrl = new GridFTPControlChannel(host, port);
    	ctrl.open();
    	ctrl.authenticate(cred);
    	Command c = new Command("SITE CHMOD 755 " + filePath);
    	Reply r = ctrl.execute(c);
    	ctrl.close();
    }
    

    /**
     * Open connection, authenticate, 
     * set transfer options
     */
    public GridFTPClient initClient(String host, int port, 
    		GSSCredential cred) throws Exception {
    	
    	// Use standard value for port unless it is specified. 
    	if(port < 0) 
    		port = 2811; 

        GridFTPClient client = new GridFTPClient(host, port);

        client.authenticate(cred); // Must be done first of all
        client.setProtectionBufferSize(65536); // TODO: Find a good number for this one
        client.setType(GridFTPSession.TYPE_IMAGE); // type of transfer
        client.setMode(GridFTPSession.MODE_EBLOCK); // transfermode
        client.setDataChannelAuthentication(DataChannelAuthentication.SELF);
        client.setDataChannelProtection(GridFTPSession.PROTECTION_SAFE);

        return client;
    }

    
    /**
     * Transfer a file from local disc to destination
     * If the executable argument is true, the file will be made executable
     * by issuing the (non standard) SITE CHMOD command. 
     * @param destination
     * @param fileName
     * @param executable
     * @throws Exception
     */
    public void transfer(String destination, 
    		String fileName, 
			boolean executable) throws FileNotFoundException, Exception {
    	
    	// Split destination in host, port and path (to file) 
    	URI destinationURI = new URI(destination); 
    	String host = destinationURI.getHost(); 
    	int port = destinationURI.getPort(); 
    	String path = destinationURI.getPath(); 
    	
    	// Open file, this will throw an exception unless file is readable. 
    	RandomAccessFile file = new RandomAccessFile(fileName, "r"); 
    	file.close(); 
    	
    	// Get delegated credential
        GSSCredential cred = null; 
        GlobusCredential gCred = JobSubmissionUtils.getDelegatedCredential(); 
        cred = new GlobusGSSCredentialImpl(gCred, 
                GSSCredential.DEFAULT_LIFETIME); 
    	
        // Set up client, transfer file and close client. 
        GridFTPClient client = null; 
        
    	logger.debug("Transferring file [" + path + "] to host " + host 
    			+ " using port " + port); 
        
        client = initClient(host, port, cred);      
        client.put(new File(fileName), path, false);
        client.close();

    	// Set executable permissions on file
        if(executable) {
        	makeFileExecutable(host, port, path, cred); 
        }
    }
    
    /**
     * Similar to transfer method, but this one performs third party transfers. 
     * @param destination
     * @param source
     * @param executable
     * @throws Exception
     */
    public void thirdPartyTransfer(String destination, 
    		String source, 
			boolean executable) throws Exception {
    	
    	URI destURI = new URI(destination); 
    	String destHost = destURI.getHost(); 
    	String destPath = destURI.getPath(); 
    	int destPort = destURI.getPort(); 
    	
    	URI sourceURI = new URI(source); 
    	String sourceHost = sourceURI.getHost(); 
    	String sourcePath = sourceURI.getPath(); 
    	int sourcePort = sourceURI.getPort(); 
    	
    	// Get delegated credential
        GSSCredential cred = null; 
        GlobusCredential gCred = JobSubmissionUtils.getDelegatedCredential(); 
        cred = new GlobusGSSCredentialImpl(gCred, 
                GSSCredential.DEFAULT_LIFETIME); 
        
    	logger.debug("Transferring file [" + sourcePath + "] from host " + sourceHost 
    			+ " using port " + sourcePort); 
    	logger.debug("To " + destPath + " on host " + destHost + 
    			" using port " + destPort); 
    	
    	GridFTPClient sourceClient = initClient(sourceHost, sourcePort, cred);
    	GridFTPClient destClient = initClient(destHost, destPort, cred); 

    	// Transfer file
    	MarkerListener listener = new DummyMarkerListener();
    	sourceClient.transfer(sourcePath, destClient,
    			destPath, false, listener);
    	
    	if(executable) {
    		makeFileExecutable(destHost, destPort, destPath, cred); 
    	}
    }
    
}
