package se.umu.jobsubmission.impl.dispatcher; 

import java.net.URI;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.globus.gsi.GlobusCredential;
import org.globus.mds.glue.GLUECERPType;
import org.globus.rsl.AbstractRslNode;
import org.globus.rsl.NameOpValue;
import org.globus.rsl.RslEvaluationException;
import org.globus.rsl.Value;
import org.nordugrid.job.gridftp.ARCGridFTPJob;
import org.nordugrid.job.gridftp.ARCGridFTPJobException;

import se.umu.jobsubmission.impl.datamanager.DataHelper;
import se.umu.reservation.ReservationInfo;
import se.umu.wsag.util.JobSubmissionUtils;


/*
   ARCDispatcher - submits jobs to ARCResources. 
   method requestJob takes an Object (contact of resource), 
   an jsdl job description (must be converted to xRSL before submission) 
   and an Object (reservation) 

   The xRSL job description is manipulated a bit to ensure that
   the GridFTP server handles is correctly. 

   */

public class ARCDispatcher implements Dispatcher {

    private static Logger logger = Logger.getLogger(ARCDispatcher.class); 

    private jsdl2xrsl descriptionTranslator; 

    public ARCDispatcher() {
        descriptionTranslator = new jsdl2xrsl(); 
    }
    
    /**
     * Get a list of local input files, and where they should be stored. 
     * @param rsl
     * @param clientURL
     * @throws Exception
     */   
    public List getLocalInputFiles(AbstractRslNode rsl) throws Exception {
    	
        NameOpValue inputfiles = rsl.getParam("inputfiles"); 

        // Values are stored as list (pair) of lists 
        if(inputfiles != null) {
        	
            List values = inputfiles.getValuesAsStrings(true); 
            Iterator iter = values.iterator(); 
            
            List removedFiles = new Vector(); 

            List result = new LinkedList(); 
            
            // Extract files stored on client host.
            // Return these files as a list of lists. 
            while(iter.hasNext()) {
                List l; 
                try {
                    l = (List) iter.next(); 
                }
                catch (ClassCastException e) {
                    throw new Exception("inputfiles must be a list of lists!"); 
                }
                
                // Format is (localURL remoteURL), e.g.:
                // N.B: localURL is local on the resource, 
                //      remoteURL is local to the client host. 
                String resourceName = (String) l.get(0); 
                String clientName = (String) l.get(1); 
                
                String scheme = null; 
                String path = null; 
                try {
                	URI clientURI = new URI(clientName); 
                	scheme = clientURI.getScheme(); 
                	path = clientURI.getPath(); 
                } catch (Exception e) {
                	logger.debug("Ignoring file " + clientName 
                			+ " due to malformed URL"); 
                	continue; 
                	// malformed URL... 
                }
                
                // boolean removeLater = false;                 
                
                // 3 possible ways to express a local file                
                // i)   (sdFileName file://.../clientFileName)
                // ii)  (sdFileName /path/to/client/file/name) 
                // iii) (sdFileName "")    (same file name on client) 
                
                // iii) 
                if(clientName.equals(resourceName)) {
                	List list = new Vector(); 
                	list.add(resourceName); 
                	list.add(clientName); 
                	result.add(list); 
                }
                // ii
                // Schema (e.g., gsiftp: is missing, this is a local file
                else if(scheme== null) {
                	List list = new Vector(); 
                	list.add(resourceName); 
                	list.add(clientName); 
                	result.add(list); 
                }
                // i
                // Scheme exist, but is file:// 
                else if("file".equals(scheme)) {
                	if(path == null || path.length() == 0) 
                		path = clientName.substring("file://".length()); 
                	List list = new Vector(); 
                	list.add(resourceName); 
                	list.add(path); 
                	result.add(list); 
                } 
 
                /*
                if(removeLater) {
                	List value = new Vector(); 
                	value.add(new Value(resourceName)); 
                	value.add(new Value(clientName)); 
                	removedFiles.add(value); 
                }*/
            }
            
            return result; 
            
            
            /*
             * TODO: Does the NG-UI remove local input files from the 
             * job description? 
             * 
            // Now, remove all files that were scheduled for transfer 
            // by the dispatcher. 
            
            for(int i = 0; i < addedFiles.size(); i++) {
            	inputfiles.add((List)addedFiles.get(i));
            }
            for(int i = 0; i < removedFiles.size(); i++) {
            	inputfiles.remove(removedFiles.get(i)); 
            }
            if(inputfiles.getValues().size() == 0) {
            	NameOpValue rel = rsl.removeParam("inputfiles");  
            	if(rel != null)
            		logger.debug("removed " + rel.getAttribute()); 
            }
            */
        }
        
        return null; 
    }

    // Adds the rsl submitted to the user to the rsl 
    // can be used for resubmitting the job. 
    public void addUserRsl(AbstractRslNode rsl) throws Exception {

        AbstractRslNode userRsl; 
        try {
           userRsl = rsl.evaluate(); 
        } catch (RslEvaluationException e) {
            logger.error("Error copying rsl"); 
            throw e; 
        }

        NameOpValue clientxrsl = new NameOpValue("clientxrsl", 
                NameOpValue.EQ, userRsl.toRSL(true)); 
        rsl.add(clientxrsl); 

    }

    /*
    For an xrsl, the value of the relation 
    (stdout=myjob.txt) 
    shout be added to the relation (outputfiles=( ... )), 
    unless there already exists an entry for stdout in outputfiles 

    TODO: Perform some typechecking of values, to ensure that 
    they are of correct object types, see xrsl2jsdl.java for details. 
    
    */
    public void addIOStaging(AbstractRslNode rsl, 
            String fileRelationName, 
            String fileListRelationName) throws Exception {

        // Check if relation was defined
        NameOpValue fileRelation = rsl.getParam(fileRelationName); 
        if(fileRelation == null) 
            return; 

        // Get value of fileRelation
        List allFileRelationValues = fileRelation.getValuesAsStrings(true); 
        String fileRelationValue = (String) allFileRelationValues.get(0); 

        logger.debug("Adding " + fileRelationValue + " to relation " 
                + fileListRelationName); 

        // Check if fileListRelation exists. 
        NameOpValue fileListRelation = rsl.getParam(fileListRelationName); 

        // Add fileListRelation to rsl, and add fileRelation to 
        // fileListRelation
        if(fileListRelation == null) {
            logger.debug("Adding new relation, " + fileListRelationName); 
            List newValues = new LinkedList(); 
            // Name "" is just a dummy, but ensures compliance with NG UI
            newValues.add(new Value(fileRelationValue)); 
            newValues.add(new Value("")); 
            NameOpValue newRelation = 
                new NameOpValue(fileListRelationName, NameOpValue.EQ); 
            newRelation.add(newValues); 
            rsl.add(newRelation); 
            return; 
        }

        // fileListRelation did exist, check if it contains
        // entry for the value of fileRelation
        List fileListRelationValues = 
            fileListRelation.getValuesAsStrings(true); 
        Iterator iter = fileListRelationValues.iterator(); 
        while(iter.hasNext()) { 
            List pair; 
            try {
                pair = (List) iter.next(); 
            } catch (ClassCastException e) {
                throw new Exception("Relation " + fileListRelationName 
                        + " must be a list of lists"); 
            }
            String localName = (String) pair.get(0); 
            if(localName.equals(fileRelationValue)) 
                return; 
        }
        
        // Entry for fileRelation is missing in fileListRelation, add it
        List newEntry = new LinkedList(); 
        // The name "" ensures that file is transferred back to used host by UI
        newEntry.add(new Value(fileRelationValue)); 
        newEntry.add(new Value("")); 
        fileListRelation.add(newEntry); 

    }

    // If executable does not start with '/', it is a local file which
    // must be staged, and added to the "executables" relation
    public void fixExecutable(AbstractRslNode rsl) throws Exception {

        NameOpValue executableRelation = rsl.getParam("executable"); 
        if(executableRelation == null) 
            throw new Exception("xRSL must contain \"executable\" relation"); 

        List executableValues = executableRelation.getValuesAsStrings(true); 
        String executableValue = (String) executableValues.get(0); 

        // Add (local file) executable to list of files that will be 
        // made executable (chmod +x) in sessiondir
        if(executableValue.charAt(0) != '/') {

            // Add files to list of files to be staged
            addIOStaging(rsl, "executable", "inputfiles"); 

            NameOpValue executablesRelation = rsl.getParam("executables"); 
            if(executablesRelation == null) {
                NameOpValue newRelation = 
                    new NameOpValue ("executables", NameOpValue.EQ, 
                            executableValue); 
                rsl.add(newRelation); 
            }
            else {
                executablesRelation.add(executableValue); 
            }
        }

        // And now, for my final trick...
        // Replace executable with /bin/echo, and place the old executable first
        // in 'arguments' (the result of `echo arguments` 
        // executed on the remote host) 
        
        // Get argument relation
        NameOpValue argumentRelation = rsl.getParam("arguments"); 

        // No args, just add executable as only argument
        if(argumentRelation == null) {
            argumentRelation = new NameOpValue("arguments", 
                    NameOpValue.EQ, executableValue); 
            rsl.add(argumentRelation); 
        }
        // Arguments exist, add executable as first argument
        else {
            // Old arguments (type String) 
            List argumentValues = argumentRelation.getValuesAsStrings(true); 
            argumentRelation.clear(); 
            argumentRelation.add(executableValue); 
            Iterator iter = argumentValues.iterator(); 
            while(iter.hasNext()) 
                argumentRelation.add((String)iter.next()); 
            
        }

        executableRelation.clear(); 
        executableRelation.add("/bin/echo"); 
    }

    // Rename stdin to sstdin
    public void fixStdin(AbstractRslNode rsl) {

        NameOpValue stdinRelation = rsl.removeParam("stdin"); 
        if(stdinRelation == null) 
            return; 

        List stdinValues = stdinRelation.getValuesAsStrings(true); 
        String stdinValue = (String) stdinValues.get(0); 
        NameOpValue sstdinRelation = new NameOpValue("sstdin", NameOpValue.EQ,
               stdinValue);  
        rsl.add(sstdinRelation); 

    }

    // Add queue name (FIXME: only uses default queue ATM)
    // This should be fixed in ARCInformationFinder 
    // and ARCInformationConverter, not here...
    public void fixQueue(AbstractRslNode rsl, GLUECERPType resourceInfo) {

        String queueName = resourceInfo.getComputingElement(0).getName(); 

        NameOpValue queueRelation = new NameOpValue("queue", 
        		NameOpValue.EQ, 
                queueName); 
        rsl.add(queueRelation); 

    }
    
    // Adds default cputime (5 minutes) unless it is specified.     
    public void fixCpuTime(AbstractRslNode rsl) {
    	
    	NameOpValue cpuTimeRel = rsl.getParam("cputime"); 
    	if(cpuTimeRel == null) {
    		logger.debug("\"cputime\" not specified, assuming " 
    				+  " minutes"); 
    		cpuTimeRel = new NameOpValue("cputime", NameOpValue.EQ, 
    				"" + Dispatcher.DEFAULT_CPU_TIME); 
    		rsl.add(cpuTimeRel); 
    	}
    	
    }
 
    /**
     * Adds hostname of client submitting job to the RSL. 
     * @param rsl
     * @param clientURL
     */
    public void fixHostName(AbstractRslNode rsl, String clientURL) {
    	
    	logger.debug("Adding \"hostname\" = " + clientURL); 
    	NameOpValue hostNameRelation = new NameOpValue("hostname", 
    			NameOpValue.EQ, 
				clientURL); 
    	rsl.add(hostNameRelation); 			
    	
    }
    
    // Add name of client software used to submit job. 
    // TODO: Find a suitable name for this piece of software :-)
    public void fixClientSoftware(AbstractRslNode rsl) {    	
    	NameOpValue clientSoftwareRelation = new NameOpValue("clientsoftware", 
    			NameOpValue.EQ, 
				"JobSubmissionService-0.1"); 
    	rsl.add(clientSoftwareRelation);     	
    }

    // Performs the required manipulations to the rsl. 
    // Add a copy of the submitted rsl in the clientxrsl relation, 
    // adds (action=request) and (savestate=yes) to rsl, 
    // add files to be staged in/out to inputfiles/outputfiles
    // renames (stdin=*) to (sstdin=*). 
    // Moved command from executable to arguments and set
    // executable to /bin/echo
    public void prepareRsl(AbstractRslNode rsl, 
    		GLUECERPType resourceInfo, 
			String clientURL) throws Exception {

        addUserRsl(rsl); 

        NameOpValue action = new NameOpValue("action", 
        		NameOpValue.EQ, 
				"request"); 
        rsl.add(action); 
        NameOpValue savestate =  new NameOpValue("savestate", 
        		NameOpValue.EQ, 
				"yes"); 
        rsl.add(savestate); 

        // Add relations (if not already defined), 
        // to ensure that files are staged in/out. 
        addIOStaging(rsl, "stdout", "outputfiles"); 
        addIOStaging(rsl, "stderr", "outputfiles"); 
        addIOStaging(rsl, "gmlog", "outputfiles"); 
        
        // TODO
        // TODO: 
        // TODO: If file staging is moved to data manager, does 
        // this still work -> No. 
        // Consequences of this??? 
        addIOStaging(rsl, "stdin", "inputfiles"); 

        fixStdin(rsl); 
        fixExecutable(rsl); 
        fixQueue(rsl, resourceInfo); 
        fixCpuTime(rsl); 
        fixHostName(rsl, clientURL);
        fixClientSoftware(rsl); 

    }
    
    
    /**
     * Stage input files for the job. 
     * Initially, try direct transfer. 
     * Should that fail, use third party transfer. 
     * Should that fail as well, the job fails... 
     * 
     * TODO: Do we need to know which file to make executable (the executable)?  
     * 
     * @param job
     * @param files
     * @param clientURL
     * @throws Exception
     */
    public void stageLocalInputFiles(String resourceURI, 
    		ARCGridFTPJob job, 
    		List files, 
			String clientURL) throws Exception {
    	
    	// Now, try to stage any input files. 
        // First directly, next via a third party transfer. 
        if(files != null) {
        	            	
        	// Direct transfer
        	List failedTransfers = new Vector(); 
        	List transfer = null; 
        	String filePath = null; 
        	String jobID = job.Id(); 
        	DataHelper dataHelper = new DataHelper(); 
        	for(int i = 0; i < files.size(); i++) {
        		try { 
        			transfer = (List) files.get(i);
        			String fileName = (String) transfer.get(0); 
        			filePath = (String) transfer.get(1); 
        			String remotePath = resourceURI + "/" + jobID + "/" + fileName; 
        			
        			dataHelper.transfer(remotePath, filePath, false); 
        		} catch (Exception e) {
        			logger.debug("Error transferring file " + filePath + " directly"); 
        			e.printStackTrace(); 
        			failedTransfers.add(transfer); 
        		}
        	}            	
        	
        	// Try third party transfers instead 
        	if(failedTransfers.size() > 0) {
        		filePath = null; 
        		String fileName = null; 
        		try {
        			for(int i = 0; i < failedTransfers.size(); i++) {
        				transfer = (List) failedTransfers.get(i); 
        				fileName = (String) transfer.get(0); 
        				filePath = (String) transfer.get(1); 
        				String remotePath = resourceURI + "/" + jobID + "/" + fileName; 
        				
        				// Now, rewrite file://path to 
        				// gsiftp://clienthost/path and try to 
        				// perform a third party transfer. 
        				URI uri = new URI(filePath); 
        				URI clientHost = new URI("gsiftp://" 
        						+ clientURL 
								+ ":2811" 
								+ uri.getPath()); 
        				dataHelper.thirdPartyTransfer(remotePath, 
        						clientHost.toString(), 
								false); 
        				
        			}
        		} catch (Exception e) {
        			// Ok, input files can not be transferred -> job failure
        			logger.debug("Can not stage input file " + filePath
        					+ " using third party transfer"); 
        			throw new ARCGridFTPJobException("Can not stage input file "
        					+ filePath); 
        		}
        	}
        }
    }

    /* 
     * Resource is a GLUECERPType
     * jobDescription is a JobDefinition_Type describing the job in JSDL
     * reservation contains the RPs from the AgreementResource. 
     */
    public Object requestJob(Object resource,    
                             Object jobDescription, 
                             Object reservation, 
							 String clientURL) throws Exception {

        GLUECERPType resourceInfo; 
        try {
            resourceInfo = (GLUECERPType) resource; 
        } catch (ClassCastException e) {
            logger.debug("Resource description of incorrect type"); 
            throw new Exception("Resource description of incorrect type", e); 
        }

        // ng-cluster-contactstring stored as 
        // Cluster:UniqueID
        String resourceURI = resourceInfo.getCluster(0).getUniqueID(); 

        ARCGridFTPJob job = null; 
        try {
        	
            job = new ARCGridFTPJob(resourceURI);
            
            // Set user's delegated credential as proxy used by job. 
            GlobusCredential cred = JobSubmissionUtils.getDelegatedCredential(); 
            job.addProxy(cred); 
            
    		// Connect to resource (using user's delegated proxy)
			job.Connect(); 
			
        } catch (ARCGridFTPJobException e) {
            logger.error("Could not create ARCGridFTPJob: " + e.getMessage()); 
            throw e; 
        }

        // Translate JSDL (back) to xRSL (if not already in xRSL format)
        // The description should always be in JSDL format, but this is useful
        // as the ARCDispatcher can be used as a stand-alone components, 
        // accepting xRSL descriptions...
        AbstractRslNode rslNode; 
        if(jobDescription instanceof AbstractRslNode) 
            rslNode = (AbstractRslNode) jobDescription; 
        else
            rslNode = (AbstractRslNode) descriptionTranslator.
                translate(jobDescription); 

        if(rslNode == null) {
            logger.error("Could not translate job description"); 
            throw new ARCGridFTPJobException
                ("could not translate job description"); 
        }
        
        // Add some commands to rsl, and 
        // add all input/output files to lists of files to stage. 
        prepareRsl(rslNode, resourceInfo, clientURL); 
        
        // Add reference to local reservation if it exists. 
        try {
        	if(reservation != null) {
        		ReservationInfo resInfo = (ReservationInfo) reservation; 
        		String localID = resInfo.getReservation().
					getLocalID().getLocalID(); 
        		
        		logger.debug("Adding (advres = " + localID + ")"); 
        		if(localID != null) {        		
        			NameOpValue advres = new NameOpValue(
        					"advres", 
							NameOpValue.EQ, 
							localID); 
        			rslNode.add(advres); 
        		}
        	}
        } catch (Exception e) {        	
        	logger.error("Error in reservation, ignoring it"); 
        }

        String rslString = rslNode.toRSL(true); 
        logger.debug("Final xRSL is: \n" + rslString);        
        
        // Submit job
        String resourceContact = null; 
        try {
            job.Submit(rslString); // , files, names); 
            resourceContact = job.Id(); 
            logger.debug("Local id is " + resourceContact);              
                        
            try { 
            	List localInputFiles = getLocalInputFiles(rslNode); 
            	if(localInputFiles != null) {
            		logger.warn("Staging input files"); 
            		stageLocalInputFiles(resourceURI, job, localInputFiles, clientURL); 
            	}
            } catch (Exception e) {
            	job.Cancel(); 
            	job.Disconnect(); 
            	throw new ARCGridFTPJobException("Error staging files: " + 
            			e.getMessage()); 
            } 
            job.Disconnect();
        }
        catch (ARCGridFTPJobException e) {
            logger.error("Job submission failed: " + e.getMessage()); 
            throw e; 
        }

        // Generate globalid, which is resourceURL/jobid
        String globalId = resourceURI; 
        if(globalId.charAt(globalId.length() -1) != '/')
            globalId += "/"; 
        globalId += resourceContact; 

        return globalId; 
    }

}
