package se.umu.jobsubmission.client; 

import java.io.BufferedReader;
import java.io.RandomAccessFile; 
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.net.InetAddress;
import java.util.Calendar;
import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.rpc.Stub;

import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.ggf.schemas.jsdl._2005._06.jsdl.JobDefinition_Type;
import org.globus.axis.util.Util;
import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.faults.BaseFaultType;

import se.umu.jobsubmission.Benchmark;
import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.JobObjectiveType;
import se.umu.jobsubmission.JobPreferences;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.JobSubmission;
import se.umu.jobsubmission.JobSubmissionServiceAddressingLocator;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.RequestJob;
import se.umu.jobsubmission.RequestJobResponse;
import se.umu.jobsubmission.SchedulingHintsType;
import se.umu.jobsubmission.impl.jobsubmissionservice.JobSubmissionConstants;
import se.umu.jobsubmission.util.JobDescriptionTranslator;
import se.umu.wsag.util.JobSubmissionUtils;


/**
 * Simple client for job submission. 
 * The client accepts four parameters, 
 * the job description (mandatory),
 * a job preferences document (optional), 
 * a list of index servers (optional), 
 * and an URL to the JSS to use (optional)
 */

public class SubmitJob {

    static {
        Util.registerTransport(); 
    }

    private String jobDescriptionFile; 
    private String jobPreferencesFile; 
    private String indexServers[]; 
    private String URI; 
    private String host; 
    private String clientHost; 
    
    private JobDescriptionTranslator jobDescriptionTranslators[]; 

    private static final String optionsDesc = 
        "OPTIONS:\n" +
        "\n"+
        "Help:\n" +
        " -help                 display help.\n" +
        " -usage                display usage.\n" +
        "Usage:\n" + 
        " -f                    Specify file that contains job description\n" + 
        " -i                    Specify IndexService to retrieve resources from\n" + 
        " -j                    Specify URL of job submission service to contact\n" +
        " -p                    Specify file that contains job preferences\n" +
        "\n"; 

    private static final String usageDesc =
        "ARGUMENTS: \n" +
        "         [options] -f <job description file name>\n" +
        "with\n" +
        "         [options]       [-i <indexService]+ \n" +
        "                         [-p <job preferences file name>]\n" +
        "                         [-j <job submission service> (default is localhost)]\n"+
        "\n";

    public SubmitJob() throws Exception {
    	
        jobDescriptionFile = null; 
        jobPreferencesFile = null; 
        // Using resolved address as 127.0.0.1 gives some problems with certificates...
        InetAddress address = InetAddress.getLocalHost();     
        host = address.getCanonicalHostName(); 
        clientHost = host; 
        
        URI = ServiceURL.getURL(
        		host, JobSubmissionConstants.SERVICE_NAME).
					toString(); 
        
        indexServers = null; 
        
    }
      
    /**
     * Parse user input
     * @param args
     * @throws Exception
     */
    public void processArguments(String args[]) throws Exception {

        List indexServersList = new Vector(); 

        if(args.length == 0) {
            System.out.println();
            System.out.println(usageDesc);
            System.out.println(optionsDesc);
            System.exit(0); 
        }

        boolean error = false; 
        for(int i = 0; i < args.length && !error; i++) {

            if (args[i].startsWith("-")) {

                if (args[i].equalsIgnoreCase("-help")) {
                    System.out.println();
                    System.out.println(usageDesc);
                    System.out.println(optionsDesc);
                    System.exit(0);
                }
                else if (args[i].equalsIgnoreCase("-usage")) {
                    System.out.println(usageDesc);
                    System.exit(0);
                }
                else if (args[i].equals("-f")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-f requires a job description file");
                        error = true; 
                    }
                    else
                        jobDescriptionFile = args[i]; 
                }
                else if (args[i].equals("-i")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-i requires a index service URI");
                        error = true; 
                    }
                    else 
                        indexServersList.add(args[i]); 
                }
                else if (args[i].equals("-j")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-j requires a job submission service URI");
                        error = true; 
                    }
                    else  {
                        host = args[i]; 
                
                        URI = ServiceURL.getURL(
                        		host, 
								JobSubmissionConstants.SERVICE_NAME).toString();
                    }
							
                }
                else if (args[i].equals("-p")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-p requires a job preferences file");
                        error = true; 
                    }
                    else
                        jobPreferencesFile = args[i]; 
                }
                else {
                    System.err.println("Unknown argument: " + args[i]);
                    error = true; 
                }
            }
            else {
                System.err.println("Unknown argument: " + args[i]);
                error = true; 
            }
        }
        if(error) {
            System.exit(1); 
        }

        int noIndexServers = indexServersList.size(); 

        if(noIndexServers > 0) {
            indexServers = new String[noIndexServers]; 
            for(int i = 0; i < noIndexServers; i++) 
                indexServers[i] = (String) indexServersList.get(i); 
        }
    }
    
    /**
     * Returns the list of job description translators as specified
     * in the client configuration file
     * @param clientConfig
     * @return
     */
    public JobDescriptionTranslator[] getConverter(String clientConfig) {
    	
    	try {
    		BufferedReader in = new BufferedReader(new FileReader(clientConfig));
    		List classNames = new Vector(); 
    		while(in.ready()) {
    			String className = in.readLine().trim();
    			classNames.add(className); 
    		}
    		in.close();
    		JobDescriptionTranslator converters[] = 
    			new JobDescriptionTranslator[classNames.size()]; 
    		
    		for(int i = 0; i < classNames.size(); i++) {
    			converters[i] = (JobDescriptionTranslator) 
				Class.forName((String)classNames.get(i)).newInstance(); 
    		}
    	
    		return converters; 
    	} catch (Exception e) {
    		System.err.println("Can not create converter:"); 
    		return null; 
    	}
    }

    /**
     * Convert job description to JSDL, 
     * the client config file determines which converter to use. 
     */
    public JobDefinition_Type convertJobDescription(String fileName) 
    throws FileNotFoundException, Exception {
    	
    	String configFile = "/etc/jobsubmission/clientConfig"; 
    	String globusLocation = ContainerConfig.getGlobusLocation(); 
    	String converterPath = globusLocation + configFile; 
    	JobDescriptionTranslator converters[] = getConverter(converterPath); 
    	
    	// Try first for file existence
    	try {
    		RandomAccessFile f = new RandomAccessFile(fileName, "r");
    		f.close(); 
    	} catch (FileNotFoundException e) {
    		throw new FileNotFoundException ("Job description file " + fileName 
    				+ " not found"); 
    	}
    	
    	for(int i = 0; i < converters.length; i++) {
    		try {
    			JobDefinition_Type jobDef = (JobDefinition_Type) converters[i].
						translate(fileName); 
    			System.out.println("Translated job description " + fileName + 
    					" using " + converters[i].getClass().getName()); 
    			return jobDef; 
    		} catch (Exception e) {
    			System.out.println("Could not use " + converters[i].getClass().getName()
    					+ " to convert job description " + fileName); 
    		}
    	}
    	throw new Exception("No valid converter exists for " + fileName); 
    	
    }
    
    /**
     * Locate JobSubmissionService, convert job description
     * and submit it to service. Include any user preferences document
     * and/or non-default index server(s) in the job request. 
     * @throws Exception
     */
    public void submitJob() throws FileNotFoundException, Exception {

        System.out.println("Submitting job using:\n" + URI);

        System.out.println("Converting job description to JSDL");
        if(jobDescriptionFile == null) {
        	System.err.println("The -p flag (JobDescription) must be specified"); 
        	throw new 
				Exception("The -p flag (JobDescription) must be specified"); 
        }        
        
        JobDefinition_Type jobDesc = convertJobDescription(jobDescriptionFile);         

        System.out.println("Locating job submission service"); 
        // Locate service
        JobSubmissionServiceAddressingLocator locator = 
            new JobSubmissionServiceAddressingLocator(); 
                        
        EndpointReferenceType endpoint = new EndpointReferenceType();
        endpoint.setAddress(new Address(URI));
        JobSubmission jobsubmissionservice = 
            locator.getJobSubmissionPort(endpoint);        
        
        // Client security descriptor file 
        String clientDesc = "etc/jobsubmission/client-security-config.xml"; 
        // Set descriptor on Stub 
        ((Stub)jobsubmissionservice)._setProperty(
        		Constants.CLIENT_DESCRIPTOR_FILE, 
				clientDesc);
                
        // Build job request message
        RequestJob request = new RequestJob(); 
        JobRequest requestParameters = new JobRequest(); 

        // Index servers
        requestParameters.setIndexServerURIs(this.indexServers); 
        
        // JobDescription
        requestParameters.setJobDescription(jobDesc); 
                
        // Jobpreferences file, deserialize it first. 
        if(this.jobPreferencesFile != null) {
        	
        	try {
        		JobPreferences preferences = null; 
        		QName preferencesQName = JobPreferences.getTypeDesc().getXmlType(); 
        		FileInputStream input = new FileInputStream(this.jobPreferencesFile); 
        		preferences = (JobPreferences) JobSubmissionUtils.deserialize( 
        				input, preferencesQName, JobPreferences.class);  
        		
        		printPreferences(preferences); 
        		
        		requestParameters.setJobPreferences(preferences); 
        	} catch (FileNotFoundException e) {
        		System.err.println("Job preferences file " + 
        				this.jobPreferencesFile + 
						" not found, no job preferences will be used"); 
        	}
        	catch (Exception e) {
        		System.err.println("Could not read job preferences from file "
        				+ this.jobPreferencesFile); 
        	}
        }
        
        // Add clientHost       
        requestParameters.setClientURL(this.clientHost);
    
        request.setRequest(requestParameters);
        
        System.out.println("Submitting job..."); 
        RequestJobResponse response = null; 
        
        // Perform actual submission
        try {
        	response = jobsubmissionservice.requestJob(request); 
        }  
        catch (InvalidJobDescriptionFaultType f) {
        	System.err.println("Job request failed, invalid job description:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoAptResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.out.println(helper.getDescriptionAsString()); 
        } catch(JobRequestFaultType f) {
        	System.err.println("Job request failed:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        }
        catch (BaseFaultType f) {        	
        	System.err.println("General fault: "); 
        	FaultHelper helper = new FaultHelper(f);         	        	        	
        	System.err.println(helper.getDescriptionAsString());         	
        }
        
        if(response != null) {
        	System.out.println("Job submitted with jobid " 
        			+ response.getJobId()); 
        }

    }
    
    /**
     * Prints the job preferences, only used for debug purposes. 
     * @param prefs
     */
    public void printPreferences(JobPreferences prefs) {
    	    	
    	if(prefs == null) {
    		System.out.println("No preferences specified"); 
    		return; 
    	}
    	
    	SchedulingHintsType hints = prefs.getSchedulingHints(); 
    	
    	if(hints != null) {
    		System.out.println("Scheduling hints:");
    		Benchmark benchmarks[] = hints.getBenchmark(); 
    		if(benchmarks != null) {
    			int size = benchmarks.length; 
    			for(int i = 0; i < size; i++) {
    				System.out.println("Benchmark: " 
    						+ benchmarks[i].getName() + " "
							+ benchmarks[i].getResult() + " "
							+ benchmarks[i].getTime()); 
    			}
    		}
    	
    	    if(hints.getFileTransfers() != null) {
    		System.out.println("Filetransfer times : in / out: " +
    			hints.getFileTransfers().getStageInTime() +
				" / " + hints.getFileTransfers().getStageOutTime());
    	    }
    	}
    	
    	JobRequirementType reqs = prefs.getJobRequirement();
    	if(reqs != null) {
    		System.out.println("Job requirements:");    	    	
    		
    		System.out.print("EarliestStart:"); 
    		Calendar c = reqs.getEarliestAbsoluteStart(); 
    		if(c != null)    		
    			System.out.println(c.getTime());
    		else
    			System.out.println(reqs.getEarliestRelativeStart());    	
    		
    		System.out.print("LatestStart: "); 
			c = reqs.getLatestAbsoluteStart(); 
			if(c != null)
				System.out.println(c.getTime()); 
			else
				System.out.println(reqs.getLatestRelativeStart()); 
    		
			Boolean useRes = reqs.getUseReservations(); 
			if(useRes == null || useRes.booleanValue()) {
								
				Boolean flex = reqs.getFlexible();
				if(flex != null && flex.booleanValue())
					System.out.println("Flexible reservation requested");
				else 
					System.out.println("Fixed reservation requested"); 
			}
			else {
				System.out.println("No reservation requested"); 
			}
			
    		JobObjectiveType objective = reqs.getJobObjective(); 
    		if(objective.equals(JobObjectiveType.earliestJobStart)) {
    			System.out.println("Earliest possible job start requested"); 
    		}
    		else if(objective.equals(JobObjectiveType.earliestJobCompletion))
    			System.out.println("Earliest possible job completion requested"); 
    		
    		String targets[] = reqs.getSubmissionTarget(); 
    		if(targets != null) {
    			System.out.println("Possible submission targets:"); 
    			for(int i = 0; i < targets.length; i++)
    				System.out.println(targets[i]); 
    		}
    		
    	}
    	    	    	    
    }

    public static void main(String args[]) {

        try {

            SubmitJob s = new SubmitJob(); 
            s.processArguments(args); 
            s.submitJob(); 

        } catch (FileNotFoundException e) {
        	System.err.println(e.getMessage()); 
        }
        catch (Exception e) {
            System.out.println("Error submitting job:");
            e.printStackTrace(); 
        }
    }
}


