package se.umu.jobsubmission.impl.submitter; 

import java.io.FileInputStream;
import java.util.Calendar;
import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._06.jsdl.JobDefinition_Type;
import org.globus.mds.glue.GLUECERPType;
import org.globus.wsrf.utils.FaultHelper; 

import org.oasis.wsrf.faults.BaseFaultType; 

import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.JobPreferences;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.SubmittedJob;
import se.umu.jobsubmission.SubmitterConfiguration;
import se.umu.jobsubmission.impl.broker.Broker;
import se.umu.jobsubmission.impl.datamanager.DataManager;
import se.umu.jobsubmission.impl.datamanager.DefaultDataManager;
import se.umu.jobsubmission.impl.dispatcher.Dispatcher;
import se.umu.jobsubmission.impl.informationfinder.InformationFinder;
import se.umu.jobsubmission.impl.informationfinder.InformationFinderFactory;
import se.umu.jobsubmission.impl.informationfinder.ResourceInformationCache;
import se.umu.jobsubmission.util.ReservationUtils;
import se.umu.reservation.ReservationInfo;
import se.umu.wsag.util.JobSubmissionUtils;

/*
   The sumbmitter is responsible for coordinating the job submission process. 
   It uses a informationFinder, the Broker, the Dispatcher an the DataManager. 
   
   An outline of Submitter.submitJob(jobDesc): 

   Broker.verify(jobDesc);  

   glueInfo = resources = informationFinder.getResourceInfo(indexServers); 

   List possibleResources = Broker.filter(glueInfo, jobDesc); 
   List rankedResources = Broker.predict(possibleResources, jobDesc); 

   boolean done = false; 
   String jobContact; 
   Resource resource; 
   while(!done) {
       try {
           done = true; 
           resource = Broker.select(rankedResources, jobDesc);
           DataManager.stageIn(jobDesc, resource); 
           jobContact = Dispatcher.requestJob(resource, jobDesc, 
               resource.hasReservation()? resource.getReservation() : null); 
       } catch (Exception e) {
           Broker.discardResource(resource); 
           done = false; 
       }
   }

   return jobContact; 
 */

public class Submitter {

    private static Logger logger = Logger.getLogger(Submitter.class); 

    private InformationFinderFactory informationFinderFactory; 
    //private String dispatcherClassName; 
    private String[] dispatcherClasses; 
    //private String dataManagerClassName; 
    private String[] dataManagerClasses; 
    private int numberOfThreads; 
    
    private static final int DEFAULT_CONNECTION_TIMEOUT = 10000; 
    private static final int DEFAULT_THREAD_POOL_SIZE = 4; 

    private ResourceInformationCache resourceInformationCache; 

    /* Set up submitter from config file. */
    public Submitter(String configFile) throws Exception {
        logger.debug("Configuring submitter from file: " + configFile); 
        this.resourceInformationCache = new ResourceInformationCache(); 
        readConfig(configFile); 
    }

    /* Config file determines which dispatchers and infofinders to use. 
       Furthermore, the finders, fetchers and converters to use in the
       infofinder are configured as well as timeouts, threadpoolsize
       and default index server for the information finder is determined. 
       Each submission uses a separate informationfinder and dispatcher
       in order to avoid concurrency issues. */
    public void readConfig(String configFile) throws Exception {

        SubmitterConfiguration config = null;
        
        QName configQName = SubmitterConfiguration.getTypeDesc().getXmlType(); 
        FileInputStream input = new FileInputStream(configFile); 
        config = (SubmitterConfiguration) JobSubmissionUtils.deserialize(
        		input, configQName, SubmitterConfiguration.class); 
        
		this.dispatcherClasses = config.getDispatcherClass(); 
		this.dataManagerClasses = config.getDataManagerClass(); 

        String finderClasses[] = config.getResourceFinderClass(); 
        String fetcherClasses[] = config.getInformationFetcherClass(); 
        String converterClasses[] = config.getInformationConverterClass(); 
        int timeout = config.getConnectionTimeout().intValue(); 
        if(timeout == 0) {
            logger.info("No timeout specified, using default (" + 
            		Submitter.DEFAULT_CONNECTION_TIMEOUT + ")"); 
            timeout = Submitter.DEFAULT_CONNECTION_TIMEOUT; 
        }
        int finderThreads = config.getThreadPoolSize().intValue(); 
        if(finderThreads == 0) {
            logger.info("Number of threads for specified, using default (" + 
            		Submitter.DEFAULT_THREAD_POOL_SIZE + ")"); 
            finderThreads = Submitter.DEFAULT_THREAD_POOL_SIZE; 
        }
        this.numberOfThreads = finderThreads; 

        String indexServers[] = config.getDefaultIndexServer(); 
        List indexServersList = new Vector(); 
        for(int i = 0; i < indexServers.length; i++) 
            indexServersList.add(indexServers[i]); 
        
        informationFinderFactory = new InformationFinderFactory(finderThreads, 
                finderClasses, 
                fetcherClasses, 
                converterClasses, 
                timeout, 
                indexServersList, 
                this.resourceInformationCache); 
    }

    /**
      Returns an array with instances of each valid Dispatcher. */
    public Dispatcher[] getDispatchers() throws Exception {
    	List objects = new Vector(); 
    	for(int i = 0; i < dispatcherClasses.length; i++) {
    		try {
    			objects.add((Dispatcher) Class.forName(
    					this.dispatcherClasses[i]).newInstance()); 
    		} catch (Exception e) {
    			logger.error("Error creating dispatcher of class: "
    					+ this.dispatcherClasses[i]); 
    		}
    	}
    	Dispatcher[] dispatchers = 
    		new Dispatcher[objects.size()]; 
    	for(int i = 0; i < objects.size(); i++) {
    		dispatchers[i] = (Dispatcher) objects.get(i); 
    	}
    	
    	return dispatchers; 
    }

    /**
      Return an array with instances of each valid DataManager. */
    public DataManager[] getDataManagers() throws Exception {
    	List managers = new Vector(); 
    	for(int i = 0; i < dataManagerClasses.length; i++) {
    		try {
    			managers.add((DataManager) Class.forName(
    					this.dataManagerClasses[i]).newInstance()); 
    		} catch (Exception e) {
    			logger.error("Error creating datamanager of class: "
    					+ this.dataManagerClasses[i]); 
    		}
    	}
    	DataManager dataManagers[] = 
    		new DataManager[managers.size() + 1]; 
    	for(int i = 0; i < managers.size(); i++) {
    		dataManagers[i] = (DataManager) managers.get(i); 
    	}
        dataManagers[dataManagers.length - 1] = new DefaultDataManager(); 
        return dataManagers; 
    }

    /**
     * Ensure that start time delay is expressed as an absolute time, 
     * convert it to absolute time if user specified a relative time.  
     * 
     * Do this for both earliest and latest start. 
     */
    public JobPreferences fixPreferences(JobPreferences pref) {

        if(pref == null)
            return null;    		

        JobRequirementType jobRequirement = pref.getJobRequirement(); 

        if(jobRequirement == null)
            return pref;

        // EarliestStart
        Calendar absolute = jobRequirement.getEarliestAbsoluteStart(); 
        Integer relative = jobRequirement.getEarliestRelativeStart(); 
        if(absolute == null) {
            logger.debug("No absolute earliest start time requested"); 
            if(relative != null) {
                logger.debug("Relative earliest start time requested-> swap"); 
                Calendar now = Calendar.getInstance();
                now.add(Calendar.MINUTE, relative.intValue()); 
                jobRequirement.setEarliestRelativeStart(new Integer(0));
                jobRequirement.setEarliestAbsoluteStart(now); 
            } 
            else 
                logger.debug("No earliest job start time requested"); 
        }

        // LatestStart
        absolute = jobRequirement.getLatestAbsoluteStart(); 
        relative = jobRequirement.getLatestRelativeStart();  
        if(absolute == null) {
            if(relative != null) {
                logger.debug("Relative latest start requested -> swap"); 
                Calendar now = Calendar.getInstance(); 
                now.add(Calendar.MINUTE, relative.intValue()); 
                jobRequirement.setLatestRelativeStart(new Integer(0)); 
                jobRequirement.setLatestAbsoluteStart(now); 					
            }
            else 
                logger.debug("No latest job start time requested"); 
        }

        return pref;

    }

    /**
      Use the informationfinder to retrieve an updated list 
      of the resources to consider when submitting job. */
    public List getResources(InformationFinder informationFinder, 
    		JobDefinition_Type jobDesc, 
            String indexServerURIs[], 
            JobPreferences jobPreferences, 
            Broker broker) throws BaseFaultType {

        String resourceURIs[] = null; 
        try {
            resourceURIs = jobPreferences.getJobRequirement().
                getSubmissionTarget(); 
        } catch (Exception e) {
            resourceURIs = null; 
        }
        // Find Grid resources. 
        logger.debug("Finding resources..."); 				
        List resources = null; 
        List indexServers = new Vector(); 

        // Discovery, and query resources. 
        if(resourceURIs == null) {
            logger.debug("Performing resource discovery"); 
            if(indexServerURIs == null || indexServerURIs.length == 0)
                indexServers = null; 
            else {
                int noIndexServers = indexServerURIs.length; 
                for(int i = 0; i < noIndexServers; i++) 
                    indexServers.add(indexServerURIs[i]); 
            }

            try {
                resources = informationFinder.getResourceInfo(indexServers, 
                        true); 
            } catch(Exception e) {
                logger.error("No resource information retrieved from "); 
                for(int i = 0; i < indexServerURIs.length; i++) {
                	logger.error(indexServerURIs[i]); 
                }
                NoResourcesFoundFaultType f = new NoResourcesFoundFaultType(); 
                FaultHelper helper = new FaultHelper(f); 
                String faultString = "No resource information retrieved from:\n"; 
                for(int i = 0; i < indexServerURIs.length; i++) {
                	faultString += indexServerURIs[i]; 
                }
                helper.addDescription(faultString); 
                throw helper.getFault(); 
            }
        }
        // Resources predefined, just query them. 
        else {
            logger.debug("Using predefined resources"); 
            try {
                for(int i = 0; i < resourceURIs.length; i++)
                    indexServers.add(resourceURIs[i]); 
                resources = informationFinder.getResourceInfo(indexServers, 
                        false); 
            } catch (Exception e) {
                logger.error("No resource information retrieved from "); 
                String faultMessage = "No resource information retrieved from:\n"; 
                for(int i = 0; i < resourceURIs.length; i++) {
                    logger.error(resourceURIs[i]); 
                    faultMessage += resourceURIs[i] + "\n"; 
                }
                // throw e; 
                NoResourcesFoundFaultType f = new NoResourcesFoundFaultType(); 
                FaultHelper helper = new FaultHelper(f); 
                helper.addDescription(faultMessage); 
                throw helper.getFault(); 
            }
        }

        logger.debug("Finding resources... done"); 
        if(informationFinder != null)
            informationFinder.shutdown(); 

        // TODO: Is this code ever reached? 
        if(resources == null || resources.size() == 0) {        	
            NoResourcesFoundFaultType f = new NoResourcesFoundFaultType(); 
            FaultHelper helper = new FaultHelper(f); 
            helper.addDescription("No resources found, can not submit job."); 
            throw helper.getFault(); 
        }

        // Filter out the resources that are not usable by the job. 
        logger.debug("Filtering resources");
        List approvedResources = broker.filter(resources, jobDesc);
        logger.debug("The following resources are approved for job ");
        for(int j = 0; j < approvedResources.size(); j++) {
        	String name = ((GLUECERPType) approvedResources.get(j)).
			getCluster(0).getName(); 
        	logger.debug(name); 
        }

        if(approvedResources.size() == 0) {
            // throw new Exception("No (suitable) resources found"); 
        	NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
        	FaultHelper helper = new FaultHelper(f); 
        	helper.addDescription("No suitable resource was found for the job"); 
        	throw helper.getFault(); 
        }

        return approvedResources; 

    }

    /**
     * Performs job submission. 
     * Validates job description(s). Uses the passed index servers URIs
     * (or default index servers), to find resources. 
     * Filters and selects resources based on job requirements and 
     * user preferences. 
     * @param jobDesc Job description(s) in DOM format. 
     * @param indexServerURIs URI to custom index servers to query. 
     * @param jobPreferences Preferences of job, including e.g. benchmarks
     * and resource selections algorithm. 
     * @return Job id for submitted job. 
     * @throws Exception
     */
    public SubmittedJob submitJob(JobDefinition_Type jobDescription, 
            String indexServerURIs[], 
            JobPreferences jobPreferences, 
            String clientURL) throws Exception {

        // Create instances of broker and ReservationUtil
        Broker broker = new Broker(this.numberOfThreads); 
        ReservationUtils reservationUtils = new ReservationUtils(); 
        broker.setReservationUtils(reservationUtils); 

        // Verify that job description contains all required 
        // attributes (currently, only testing for "Executable"). 
        // This call throws InvalidJobDescriptionFaultType upon error.
        broker.verify(jobDescription); 

        // Convert formats in job preferences, including
        // convertion of relative start delays to absolute time.
        JobPreferences prefs = null; 
        if(jobPreferences != null) {
            prefs = fixPreferences(jobPreferences);        
            if(prefs == null)
                logger.error("Error fixing job preferences");
        }
        else 
            logger.debug("No job preferences specified"); 

        // Find resources, and create a list for each job 
        // containing the resources able to execute the job. 
        logger.info("Getting resources"); 
        List approvedResources = null; 
        InformationFinder informationFinder = informationFinderFactory.
            getInformationFinder(); 
       
        approvedResources = getResources(informationFinder, 
        		jobDescription, 
				indexServerURIs, 
				jobPreferences, 
				broker); 
     
        logger.info("Done getting resources."); 

        // Select resource for job and submit job. 
        List jobIds = new Vector(); 
        Object jobId = null;                

        logger.info("We have " + approvedResources.size() + " resources to select from"); 
        
        logger.info("Reordering resources"); 
        // Sort resources in decreasing order of aptness, 
        // this may include creating advance reservations.
        List orderedResources = broker.select(approvedResources,
        		jobDescription, 
				jobPreferences);
        logger.info("Done reordering resources"); 
        
        if(orderedResources.size() == 0) {
        	logger.error("No suitable resources found for job"); 
        	NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
        	FaultHelper helper = new FaultHelper(f); 
        	helper.setDescription("No suitable resources found for job"); 
        	throw helper.getFault(); 
        }
        
        GLUECERPType selectedResource = null;
        boolean done = false; 
        boolean failed = false; 
        ReservationInfo reservation = null; 
        
        SubmittedJob result = new SubmittedJob(); 
        
        Dispatcher dispatchers[] = this.getDispatchers(); 
        DataManager dataManagers[] = this.getDataManagers(); 
        
        // Check is a latest job start was specified. 
        Calendar latestStart = null; 
        boolean useLatestStart = false; 
        if(jobPreferences != null) {
        	latestStart = jobPreferences.getJobRequirement().
			getLatestAbsoluteStart(); 
        	if(latestStart != null)
        		useLatestStart = true; 
        }
        
        // Try all suitable resources in decreased order of aptness. 
        while(!done) {
        	
        	try {
        		
        		logger.debug("There remains " + orderedResources.size() 
        				+ " resources to selected from"); 
        		selectedResource = (GLUECERPType) orderedResources.remove(0);
        		
        		// No more resources to select from. 
        		if(selectedResource == null) {
        			done = true; 
        			failed = true; 
        			continue; 
        		}
        		
        		logger.debug("Submitting job to: " 
        				+  selectedResource.getCluster(0).getName()); 
        		
        		// Get reservation for job. 
        		reservation = reservationUtils.
				getReservationInfo((Object)selectedResource); 
        		
        		// ensure that latestStart is not too late...
        		if(useLatestStart) {
        			logger.debug("Checking that start is not to late..."); 
        			
        			// No reservation -> no chance job will start in time
        			if(reservation == null) {
        				logger.debug("Job start is to late!"); 
        				reservationUtils.removeUnusedReservations(); 
        				NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
        				FaultHelper helper = new FaultHelper(f); 
        				helper.addDescription("Job start is too late"); 
        				throw helper.getFault(); 
        			}
        			// If we have a reservation, ensure it starts 
        			// early enough...
        			Calendar resStart = reservation.getReservation().
					getExactStart().getExactStart(); 
        			if(resStart.after(latestStart)) {
        				logger.debug("Job start is too late!"); 
        				reservationUtils.removeUnusedReservations(); 
        				// throw new Exception("Failed to submit job");
        				NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
        				FaultHelper helper = new FaultHelper(f); 
        				helper.addDescription("Job start is too late"); 
        				throw helper.getFault(); 
        			}
        			else {
        				logger.debug("Reservation starts early enough"); 
        			}
        		}
        		
        		// Stage input files
        		boolean staged = false; 
        		for(int i = 0; i < dataManagers.length; i++) {
        			logger.info("Staging input data using: "  + dataManagers[i].getClass()); 
        			try {
        				dataManagers[i].stageIn(selectedResource, 
        						jobDescription, 
								clientURL); 
        				staged = true; 
        			} catch (Exception e) {
        				logger.error("Could not stage input files using: " 
        						+ dataManagers[i].getClass()); 
        			}
        			logger.info("Done stating input data"); 
        		}
        		if(!staged) {
        			logger.error("Could not stage input data"); 
        			throw new Exception("Failed to stage input data"); 
        		}
        		
        		boolean dispatched = false; 
        		for(int i = 0; i < dispatchers.length && !dispatched; i++) {
        			logger.info("Dispatching job using: " + 
        					dispatchers[i].getClass()); 
        			try {
        				jobId = dispatchers[i].requestJob(selectedResource, 
        						jobDescription, 
								reservation, // null -> no reservation is created.
								clientURL); 
        				dispatched = true; 
        			} catch (Exception e) {        			
        				logger.error("Could not dispatch job using: " 
        						+ dispatchers[i].getClass()); 
        			}
        		}
        		if(!dispatched) {
        			logger.error("Could not dispatch job"); 
        			throw new Exception("Could not dispatch job"); 
        		}
        		
        		logger.info("Done dispatching job"); 
        		logger.info("Confirming reservation"); 
        		
        		// TODO: Should not call this if no reservation 
        		// was created. if(reservation == null) ...
        		try {
        			reservationUtils.confirmReservation(selectedResource); 
        		} catch (Exception e) {
        			logger.debug("Could not confirm reservation / "
        					+ " no reservation to confirm"); 
        		}
        		logger.info("Done confirming reservation"); 
        		
        		done = true; 
        		
        		// Remove those reservations that are not used. 
        		logger.debug("Releasing unused reservation"); 
        		reservationUtils.removeUnusedReservations(); 
        		
        		// Update informationfinder cache. 
        		logger.warn("Updating cache"); 
        		informationFinder.getResourceInformationCache().update(
        				selectedResource, jobDescription); 					
        		
        	}
        	catch (Exception e) {
        		logger.error("Submission to " + selectedResource.
        				getCluster(0).getName() + " failed: "); 
        		
        		reservationUtils.removeReservation(selectedResource); 
        		if(orderedResources.size() == 0) {
        			failed = true; 
        			done = true; 
        		}
        	}
        }
        if(!failed) {
        	logger.info("Job submitted with jobid " + jobId); 
        	result.setLocalID(jobId.toString());
        	result.setReservationInfo(reservation); 
        	//result.setJobDescription(jobDesc); 
        	result.setJobDescription(jobDescription); 
        	return result; 
        }
        else {
        	logger.error("Failed to submit job"); 
        	JobRequestFaultType f = new JobRequestFaultType(); 
        	FaultHelper helper = new FaultHelper(f); 
        	helper.addDescription("All job submission attempts fail"); 
        	throw helper.getFault();
        	// throw new Exception("Failed to submit job");
        }
    }

    /**
     * Destroyes thread pool, shuts down open connections, etc. 
     *
     */
    public void shutdown() {
        // informationFinder.shutdown(); 
    }

}


